@file:JvmName("TransactionMethodsKt")

package ru.landgrafhomyak.serdha.api.v0.runtime

import kotlin.contracts.InvocationKind
import kotlin.contracts.contract
import kotlin.jvm.JvmName
import ru.landgrafhomyak.serdha.api.v0.LowLevelApi
import ru.landgrafhomyak.serdha.api.v0.dml.Insert
import ru.landgrafhomyak.serdha.api.v0.dml.Select
import ru.landgrafhomyak.serdha.api.v0.dml.Update


@Suppress("FunctionName")
@PublishedApi
internal inline fun <R> _safeAutoClose(onAbort: () -> Unit = {}, onSuccess: () -> Unit = {}, action: () -> R): R {
	contract {
		callsInPlace(action, InvocationKind.EXACTLY_ONCE)
		callsInPlace(onAbort, InvocationKind.AT_MOST_ONCE)
		callsInPlace(onSuccess, InvocationKind.AT_MOST_ONCE)
	}
	val ret: R
	try {
		ret = action()
	} catch (e1: Throwable) {
		try {
			onAbort()
		} catch (e2: Throwable) {
			e1.addSuppressed(e2)
		}
		throw e1
	}
	onSuccess()
	return ret
}

@Suppress("FunctionName")
@PublishedApi
internal inline fun <R> _safeAutoClose(finally: () -> Unit, action: () -> R): R {
	contract {
		callsInPlace(action, InvocationKind.EXACTLY_ONCE)
		callsInPlace(finally, InvocationKind.EXACTLY_ONCE)
	}
	return _safeAutoClose(finally, finally, action)
}

public class QueryWithoutReturnError(message: String, public val query: Any) : Error(message)

@Suppress("FunctionName")
@PublishedApi
@OptIn(LowLevelApi::class)
internal suspend inline fun <Q : Any> _wrapWithLambdas(
	compiledQuery: Any,
	queryGetter: () -> _ParametersSetter<Q, _ResultSet<Q>?>,
	parameters: (ParametersSetter<Q>) -> Unit,
	hasReturning: Boolean,
	returning: (_ResultSet<Q>) -> Unit,
	queryWithoutErrorMessage: String
) {
	contract {
		callsInPlace(queryGetter, InvocationKind.EXACTLY_ONCE)
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(returning, InvocationKind.AT_MOST_ONCE)
	}
	val paramsSetter = queryGetter()

	_safeAutoClose(onAbort = { paramsSetter._abort() }) {
		parameters(paramsSetter)
	}

	val rows = paramsSetter._execute()
	if (rows == null) {
		if (hasReturning)
			throw QueryWithoutReturnError(queryWithoutErrorMessage, compiledQuery)
		return
	}

	_safeAutoClose(finally = { rows._close() }) {
		returning(rows)
	}
}

public class ResultNotSingleError : Error()

@Suppress("FunctionName")
@PublishedApi
@OptIn(LowLevelApi::class)
internal suspend inline fun <Q : Any, R> _wrapWithLambdasSingleOrNull(
	compiledQuery: Any,
	queryGetter: () -> _ParametersSetter<Q, _ResultSet<Q>?>,
	parameters: (ParametersSetter<Q>) -> Unit,
	hasReturning: Boolean,
	returning: (Row<Q>) -> R,
	queryWithoutErrorMessage: String
): R? {
	contract {
		callsInPlace(queryGetter, InvocationKind.EXACTLY_ONCE)
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(returning, InvocationKind.AT_MOST_ONCE)
	}

	_wrapWithLambdas(
		compiledQuery, queryGetter, parameters, hasReturning,
		{ rs ->
			if (!rs._next())
				return@_wrapWithLambdasSingleOrNull null
			val ret = returning(rs)
			if (rs._next())
				throw ResultNotSingleError()
			return@_wrapWithLambdasSingleOrNull ret
		},
		queryWithoutErrorMessage
	)
	throw QueryWithoutReturnError(queryWithoutErrorMessage, compiledQuery)
}


@Suppress("FunctionName")
@PublishedApi
@OptIn(LowLevelApi::class)
internal suspend inline fun <Q : Any> _wrapWithLambdasIterate(
	compiledQuery: Any,
	queryGetter: () -> _ParametersSetter<Q, _ResultSet<Q>?>,
	parameters: (ParametersSetter<Q>) -> Unit,
	hasReturning: Boolean,
	returning: (Row<Q>) -> Unit,
	queryWithoutErrorMessage: String
) {
	contract {
		callsInPlace(queryGetter, InvocationKind.EXACTLY_ONCE)
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(returning, InvocationKind.UNKNOWN)
	}

	return _wrapWithLambdas(compiledQuery, queryGetter, parameters, hasReturning, { rs -> while (rs._next()) returning(rs) }, queryWithoutErrorMessage)
}

@Suppress("FunctionName")
@PublishedApi
@OptIn(LowLevelApi::class)
internal suspend inline fun <Q : Any, R> _wrapWithLambdasMap(
	compiledQuery: Any,
	queryGetter: () -> _ParametersSetter<Q, _ResultSet<Q>?>,
	parameters: (ParametersSetter<Q>) -> Unit,
	hasReturning: Boolean,
	returning: (Row<Q>) -> R,
	queryWithoutErrorMessage: String
): List<R> {
	contract {
		callsInPlace(queryGetter, InvocationKind.EXACTLY_ONCE)
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(returning, InvocationKind.UNKNOWN)
	}

	_wrapWithLambdas(
		compiledQuery, queryGetter, parameters, hasReturning,
		{ rs ->
			val out = ArrayList<R>()
			while (rs._next())
				out.add(returning(rs))
			return@_wrapWithLambdasMap out
		},
		queryWithoutErrorMessage
	)
	throw QueryWithoutReturnError(queryWithoutErrorMessage, compiledQuery)
}


@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> Transaction.selectSingleOrNull(
	compiledQuery: Select<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	rowsConsumer: (Row<QueryUserWrapper>) -> R
): R? {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.AT_MOST_ONCE)
	}
	return _wrapWithLambdasSingleOrNull(
		compiledQuery = compiledQuery,
		queryGetter = { this._select(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> Transaction.selectThenIterate(
	compiledQuery: Select<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit,
	rowsConsumer: (Row<QueryUserWrapper>) -> Unit
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.UNKNOWN)
	}
	_wrapWithLambdasIterate(
		compiledQuery = compiledQuery,
		queryGetter = { this._select(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> Transaction.selectThenMap(
	compiledQuery: Select<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit,
	rowsConsumer: (Row<QueryUserWrapper>) -> R
): List<R> {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.UNKNOWN)
	}
	return _wrapWithLambdasMap(
		compiledQuery = compiledQuery,
		queryGetter = { this._select(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}


@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> Transaction.insert(
	compiledQuery: Insert<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
	}
	_wrapWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._insert(compiledQuery) },
		parameters = parameters,
		hasReturning = false,
		returning = {},
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> Transaction.insertReturningSingleOrNull(
	compiledQuery: Insert<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	rowsConsumer: (Row<QueryUserWrapper>) -> R
): R? {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.AT_MOST_ONCE)
	}
	return _wrapWithLambdasSingleOrNull(
		compiledQuery = compiledQuery,
		queryGetter = { this._insert(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> Transaction.insertReturningIterate(
	compiledQuery: Insert<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit,
	rowsConsumer: (Row<QueryUserWrapper>) -> Unit
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.UNKNOWN)
	}
	_wrapWithLambdasIterate(
		compiledQuery = compiledQuery,
		queryGetter = { this._insert(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> Transaction.insertReturningMap(
	compiledQuery: Insert<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit,
	rowsConsumer: (Row<QueryUserWrapper>) -> R
): List<R> {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.UNKNOWN)
	}
	return _wrapWithLambdasMap(
		compiledQuery = compiledQuery,
		queryGetter = { this._insert(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> Transaction.update(
	compiledQuery: Update<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
	}
	_wrapWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._update(compiledQuery) },
		parameters = parameters,
		hasReturning = false,
		returning = {},
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> Transaction.updateReturningSingleOrNull(
	compiledQuery: Update<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	rowsConsumer: (Row<QueryUserWrapper>) -> R
): R? {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.AT_MOST_ONCE)
	}
	return _wrapWithLambdasSingleOrNull(
		compiledQuery = compiledQuery,
		queryGetter = { this._update(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> Transaction.updateReturningIterate(
	compiledQuery: Update<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit,
	rowsConsumer: (Row<QueryUserWrapper>) -> Unit
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.UNKNOWN)
	}
	_wrapWithLambdasIterate(
		compiledQuery = compiledQuery,
		queryGetter = { this._update(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> Transaction.updateReturningMap(
	compiledQuery: Update<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit,
	rowsConsumer: (Row<QueryUserWrapper>) -> R
): List<R> {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.UNKNOWN)
	}
	return _wrapWithLambdasMap(
		compiledQuery = compiledQuery,
		queryGetter = { this._update(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend fun <R> SynchronizedDatabase<*>.transaction(body: (Transaction) -> R):R {
	contract {
		callsInPlace(body, InvocationKind.EXACTLY_ONCE)
	}

	val tr = this._startTransaction()
	_safeAutoClose(onAbort = { tr.rollback(); tr._assertTransactionFinishedAndReleaseResources() }, onSuccess = { tr._assertTransactionFinishedAndReleaseResources() }) {
		return@transaction body(tr)
	}
}